import pytest

def pytest_addoption(parser):
    parser.addoption('--runslow', action='store_true',
                     default=False, help='run slow tests')

def pytest_collection_modifyitems(config, items):
    if config.getoption('--runslow'):
        # --runslow given in cli: do not skip slow tests
        return
    skip_slow = pytest.mark.skip(reason='need --runslow option to run')
    for item in items:
        if 'slow' in item.keywords:
            item.add_marker(skip_slow)

def pytest_runtest_teardown(item):
    if 'tmpdir' in item.funcargs:
        tmpdir = item.funcargs['tmpdir']
        if tmpdir.check():
            tmpdir.remove()



from __future__ import absolute_import

import pytest
import os
import sys
try:
    from cStringIO import StringIO as BytesIO
except ImportError:
    from io import BytesIO
import struct

prefix = '.'
for i in range(0, 3):
    if os.path.isdir(os.path.join(prefix, 'pycdlib')):
        sys.path.insert(0, prefix)
        break
    else:
        prefix = '../' + prefix

import pycdlib.isohybrid

# APMPartHeader
def test_apm_part_header_parse_initialized_twice():
    apm_part = pycdlib.isohybrid.APMPartHeader()
    apm_part.parse(b'\x50\x4d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x00'*32 + b'\x00'*32 + b'\x00\x00\x00\x00'*10 + b'\x00'*16 + b'\x00\x00\x00\x00' + b'\x00'*372)

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        apm_part.parse(b'\x50\x4d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x00'*32 + b'\x00'*32 + b'\x00\x00\x00\x00'*10 + b'\x00'*16 + b'\x00\x00\x00\x00' + b'\x00'*372)
    assert(str(excinfo.value) == 'This APMPartHeader object is already initialized')

def test_apm_part_header_parse_not_magic():
    apm_part = pycdlib.isohybrid.APMPartHeader()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInvalidISO) as excinfo:
        apm_part.parse(b'\x40\x4d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x00'*32 + b'\x00'*32 + b'\x00\x00\x00\x00'*10 + b'\x00'*16 + b'\x00\x00\x00\x00' + b'\x00'*372)
    assert(str(excinfo.value) == 'Invalid APM signature')

def test_apm_part_header_new_initialized_twice():
    apm_part = pycdlib.isohybrid.APMPartHeader()
    apm_part.new('foo', 'type', 0)

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        apm_part.new('foo', 'type', 0)
    assert(str(excinfo.value) == 'This APMPartHeader object is already initialized')

def test_apm_part_header_record_not_initialized():
    apm_part = pycdlib.isohybrid.APMPartHeader()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        apm_part.record()
    assert(str(excinfo.value) == 'This APMPartHeader object is not initialized')

# GPTPartHeader
def test_gpt_part_header_parse_initialized_twice():
    gpt_part = pycdlib.isohybrid.GPTPartHeader()
    gpt_part.parse(b'\xa2\xa0\xd0\xeb\xe5\xb9\x33\x44\x87\xc0\x68\xb6\xb7\x26\x99\xc7' + b'\x00'*16 + b'\x00'*16 + b'\x00'*8 + b'\x00'*72)

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        gpt_part.parse(b'\xa2\xa0\xd0\xeb\xe5\xb9\x33\x44\x87\xc0\x68\xb6\xb7\x26\x99\xc7' + b'\x00'*16 + b'\x00'*16 + b'\x00'*8 + b'\x00'*72)
    assert(str(excinfo.value) == 'This GPTPartHeader object is already initialized')

def test_gpt_part_header_parse_invalid_guid():
    gpt_part = pycdlib.isohybrid.GPTPartHeader()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInvalidISO) as excinfo:
        gpt_part.parse(b'\x00'*16 + b'\x00'*16 + b'\x00'*16 + b'\x00'*8 + b'\x00'*72)
    assert(str(excinfo.value) == 'Invalid Partition Type UUID')

def test_gpt_part_header_new_initialized_twice():
    gpt_part = pycdlib.isohybrid.GPTPartHeader()
    gpt_part.new(True, 'foo')

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        gpt_part.new(True, 'foo')
    assert(str(excinfo.value) == 'This GPTPartHeader object is already initialized')

def test_gpt_part_header_record_not_initialized():
    gpt_part = pycdlib.isohybrid.GPTPartHeader()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        gpt_part.record()
    assert(str(excinfo.value) == 'This GPTPartHeader object is not initialized')

# GPTHeader
def test_gpt_header_parse_initialized_twice():
    gpt_header = pycdlib.isohybrid.GPTHeader()
    gpt_header.parse(b'\x45\x46\x49\x20\x50\x41\x52\x54' + b'\x00\x00\x01\x00' + b'\x5c\x00\x00\x00' + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x00'*16 + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x00'*420)

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        gpt_header.parse(b'\x45\x46\x49\x20\x50\x41\x52\x54' + b'\x00\x00\x01\x00' + b'\x5c\x00\x00\x00' + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x00'*16 + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x00'*420)
    assert(str(excinfo.value) == 'This GPTHeader object is already initialized')

def test_gpt_header_parse_bad_gpt_sig():
    gpt_header = pycdlib.isohybrid.GPTHeader()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInvalidISO) as excinfo:
        gpt_header.parse(b'\x44\x46\x49\x20\x50\x41\x52\x54' + b'\x00\x00\x01\x00' + b'\x5c\x00\x00\x00' + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x00'*16 + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x00'*420)
    assert(str(excinfo.value) == 'Failed to find GPT signature while parsing GPT Header')

def test_gpt_header_parse_bad_gpt_rev():
    gpt_header = pycdlib.isohybrid.GPTHeader()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInvalidISO) as excinfo:
        gpt_header.parse(b'\x45\x46\x49\x20\x50\x41\x52\x54' + b'\x01\x00\x01\x00' + b'\x5c\x00\x00\x00' + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x00'*16 + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x00'*420)
    assert(str(excinfo.value) == 'Failed to find GPT revision while parsing GPT Header')

def test_gpt_header_parse_bad_header_size():
    gpt_header = pycdlib.isohybrid.GPTHeader()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInvalidISO) as excinfo:
        gpt_header.parse(b'\x45\x46\x49\x20\x50\x41\x52\x54' + b'\x00\x00\x01\x00' + b'\x5b\x00\x00\x00' + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x00'*16 + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x00'*420)
    assert(str(excinfo.value) == 'Invalid GPT Header size while parsing GPT Header')

def test_gpt_header_new_initialized_twice():
    gpt_header = pycdlib.isohybrid.GPTHeader()
    gpt_header.new(False)

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        gpt_header.new(False)
    assert(str(excinfo.value) == 'This GPTHeader object is already initialized')

def test_gpt_header_set_lbas_not_initialized():
    gpt_header = pycdlib.isohybrid.GPTHeader()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        gpt_header.set_lbas(0, 0)
    assert(str(excinfo.value) == 'This GPTHeader object is not initialized')

def test_gpt_header_set_last_usable_lba_not_initialized():
    gpt_header = pycdlib.isohybrid.GPTHeader()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        gpt_header.set_last_usable_lba(0)
    assert(str(excinfo.value) == 'This GPTHeader object is not initialized')

def test_gpt_header_record_not_initialized():
    gpt_header = pycdlib.isohybrid.GPTHeader()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        gpt_header.record(0)
    assert(str(excinfo.value) == 'This GPTHeader object is not initialized')

# GPT
def test_gpt_parse_primary_initialized_twice():
    gpt = pycdlib.isohybrid.GPT(True)
    gpt.parse_primary(b'\x00'*512 + b'\x45\x46\x49\x20\x50\x41\x52\x54' + b'\x00\x00\x01\x00' + b'\x5c\x00\x00\x00' + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x00'*16 + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x00'*420, False)

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        gpt.parse_primary(b'\x00'*512 + b'\x45\x46\x49\x20\x50\x41\x52\x54' + b'\x00\x00\x01\x00' + b'\x5c\x00\x00\x00' + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x00'*16 + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x00'*420, False)
    assert(str(excinfo.value) == 'This GPT object is already initialized')

def test_gpt_parse_primary_with_secondary_GPT():
    gpt = pycdlib.isohybrid.GPT(False)

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        gpt.parse_primary(b'\x00'*512 + b'\x45\x46\x49\x20\x50\x41\x52\x54' + b'\x00\x00\x01\x00' + b'\x5c\x00\x00\x00' + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x00'*16 + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x00'*420, False)
    assert(str(excinfo.value) == 'Cannot parse primary with a secondary GPT')

def test_gpt_parse_secondary_header_initialized_twice():
    gpt = pycdlib.isohybrid.GPT(False)
    gpt.parse_secondary_header(b'\x45\x46\x49\x20\x50\x41\x52\x54' + b'\x00\x00\x01\x00' + b'\x5c\x00\x00\x00' + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x00'*16 + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x00'*420)
    gpt.parse_secondary_partitions(b'')

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        gpt.parse_secondary_header(b'\x45\x46\x49\x20\x50\x41\x52\x54' + b'\x00\x00\x01\x00' + b'\x5c\x00\x00\x00' + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x00'*16 + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x00'*420)
    assert(str(excinfo.value) == 'This GPT object is already initialized')

def test_gpt_parse_secondary_header_with_primary_gpt():
    gpt = pycdlib.isohybrid.GPT(True)

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        gpt.parse_secondary_header(b'\x45\x46\x49\x20\x50\x41\x52\x54' + b'\x00\x00\x01\x00' + b'\x5c\x00\x00\x00' + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x00'*16 + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x00'*420)
    assert(str(excinfo.value) == 'Cannot parse secondary header with a primary GPT')

def test_gpt_parse_secondary_header_twice():
    gpt = pycdlib.isohybrid.GPT(False)
    gpt.parse_secondary_header(b'\x45\x46\x49\x20\x50\x41\x52\x54' + b'\x00\x00\x01\x00' + b'\x5c\x00\x00\x00' + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x00'*16 + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x00'*420)
    gpt.parse_secondary_partitions(b'')

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        gpt.parse_secondary_partitions(b'')
    assert(str(excinfo.value) == 'This GPT object is already initialized')

def test_gpt_new_initialized_twice():
    gpt = pycdlib.isohybrid.GPT(True)
    gpt.new(False)

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        gpt.new(False)
    assert(str(excinfo.value) == 'This GPT object is already initialized')

def test_gpt_record_not_initialized():
    gpt = pycdlib.isohybrid.GPT(True)

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        gpt.record()
    assert(str(excinfo.value) == 'This GPT object is not initialized')

# IsoHybrid
def test_isohybrid_parse_initialized_twice():
    isohybrid = pycdlib.isohybrid.IsoHybrid()
    isohybrid.parse(b'\x33\xed' + b'\x90' * 30 + b'\x00'*400 + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x80' + b'\x00'*63 + b'\x55\xaa')

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        isohybrid.parse(b'\x33\xed' + b'\x90' * 30 + b'\x00'*400 + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x80' + b'\x00'*63 + b'\x55\xaa')
    assert(str(excinfo.value) == 'This IsoHybrid object is already initialized')

def test_isohybrid_parse_invalid_size():
    isohybrid = pycdlib.isohybrid.IsoHybrid()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        isohybrid.parse(b'')
    assert(str(excinfo.value) == 'Invalid IsoHybrid MBR passed')

def test_isohybrid_parse_invalid_unused1():
    isohybrid = pycdlib.isohybrid.IsoHybrid()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInvalidISO) as excinfo:
        isohybrid.parse(b'\x33\xed' + b'\x90' * 30 + b'\x00'*400 + b'\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x80' + b'\x00'*63 + b'\x55\xaa')
    assert(str(excinfo.value) == 'Invalid IsoHybrid unused1')

def test_isohybrid_parse_invalid_unused2():
    isohybrid = pycdlib.isohybrid.IsoHybrid()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInvalidISO) as excinfo:
        isohybrid.parse(b'\x33\xed' + b'\x90' * 30 + b'\x00'*400 + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00' + b'\x80' + b'\x00'*63 + b'\x55\xaa')
    assert(str(excinfo.value) == 'Invalid IsoHybrid unused2')

def test_isohybrid_parse_no_part_entry():
    isohybrid = pycdlib.isohybrid.IsoHybrid()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInvalidISO) as excinfo:
        isohybrid.parse(b'\x33\xed' + b'\x90' * 30 + b'\x00'*400 + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x00' + b'\x00'*63 + b'\x55\xaa')
    assert(str(excinfo.value) == 'No valid partition found in IsoHybrid!')

def test_isohybrid_parse_bad_tail():
    isohybrid = pycdlib.isohybrid.IsoHybrid()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInvalidISO) as excinfo:
        isohybrid.parse(b'\x33\xed' + b'\x90' * 30 + b'\x00'*400 + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x80' + b'\x00'*63 + b'\xaa\xaa')
    assert(str(excinfo.value) == 'Invalid tail on isohybrid section')

def test_isohybrid_parse_secondary_gpt_header_not_initialized():
    isohybrid = pycdlib.isohybrid.IsoHybrid()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        isohybrid.parse_secondary_gpt_header(b'')
    assert(str(excinfo.value) == 'This IsoHybrid object is not initialized')

def test_isohybrid_parse_secondary_gpt_partitions_not_initialized():
    isohybrid = pycdlib.isohybrid.IsoHybrid()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        isohybrid.parse_secondary_gpt_partitions(b'')
    assert(str(excinfo.value) == 'This IsoHybrid object is not initialized')

def test_isohybrid_parse_clamp_geometry_sectors():
    isohybrid = pycdlib.isohybrid.IsoHybrid()

    isohybrid.parse(b'\x33\xed' + b'\x90' * 30 + b'\x00'*400 + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + b'\x80' + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00' + b'\x00'*47 + b'\x55\xaa')
    assert(isohybrid.geometry_sectors == 63)

def test_isohybrid_new_initialized_twice():
    isohybrid = pycdlib.isohybrid.IsoHybrid()
    isohybrid.new(False, False, 0, 0, 0, 1, 1, 0)

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        isohybrid.new(False, False, 0, 0, 0, 1, 1, 0)
    assert(str(excinfo.value) == 'This IsoHybrid object is already initialized')

def test_isohybrid_new_bad_sectors():
    isohybrid = pycdlib.isohybrid.IsoHybrid()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInvalidInput) as excinfo:
        isohybrid.new(False, False, 0, 0, 0, 0, 1, 0)
    assert(str(excinfo.value) == 'Geometry sectors can only be between 1 and 63, inclusive')

def test_isohybrid_new_bad_heads():
    isohybrid = pycdlib.isohybrid.IsoHybrid()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInvalidInput) as excinfo:
        isohybrid.new(False, False, 0, 0, 0, 1, 0, 0)
    assert(str(excinfo.value) == 'Geometry heads can only be between 1 and 256, inclusive')

def test_isohybrid_new_bad_mac_part_type():
    isohybrid = pycdlib.isohybrid.IsoHybrid()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInvalidInput) as excinfo:
        isohybrid.new(False, True, 0, 0, 0, 1, 1, 1)
    assert(str(excinfo.value) == 'When generating for Mac, partition type must be 0')

def test_isohybrid_record_not_initialized():
    isohybrid = pycdlib.isohybrid.IsoHybrid()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        isohybrid.record(1)
    assert(str(excinfo.value) == 'This IsoHybrid object is not initialized')

def test_isohybrid_record_padding_not_initialized():
    isohybrid = pycdlib.isohybrid.IsoHybrid()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        isohybrid.record_padding(1)
    assert(str(excinfo.value) == 'This IsoHybrid object is not initialized')

def test_isohybrid_update_rba_not_initialized():
    isohybrid = pycdlib.isohybrid.IsoHybrid()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        isohybrid.update_rba(1)
    assert(str(excinfo.value) == 'This IsoHybrid object is not initialized')

def test_isohybrid_update_efi_not_initialized():
    isohybrid = pycdlib.isohybrid.IsoHybrid()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        isohybrid.update_efi(1, 1, 1)
    assert(str(excinfo.value) == 'This IsoHybrid object is not initialized')

def test_isohybrid_update_efi_not_efi():
    isohybrid = pycdlib.isohybrid.IsoHybrid()
    isohybrid.new(False, False, 0, 0, 0, 1, 1, 0)

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        isohybrid.update_efi(1, 1, 1)
    assert(str(excinfo.value) == 'Attempted to set EFI lba on a non-EFI ISO')

def test_isohybrid_update_mac_not_initialized():
    isohybrid = pycdlib.isohybrid.IsoHybrid()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        isohybrid.update_mac(1, 1)
    assert(str(excinfo.value) == 'This IsoHybrid object is not initialized')

def test_isohybrid_update_mac_not_mac():
    isohybrid = pycdlib.isohybrid.IsoHybrid()
    isohybrid.new(False, False, 0, 0, 0, 1, 1, 0)

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        isohybrid.update_mac(1, 1)
    assert(str(excinfo.value) == 'Attempted to set Mac lba on a non-Mac ISO')



# Copyright (C) 2018-2019  Chris Lalancette <clalancette@gmail.com>

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

'''
PyCdlib Inode class.
'''

from __future__ import absolute_import

from pycdlib import pycdlibexception

# For mypy annotations
if False:  # pylint: disable=using-constant-test
    from typing import BinaryIO, List, Optional, Tuple, Union  # NOQA pylint: disable=unused-import
    # NOTE: this import has to be here to avoid circular deps
    from pycdlib import dr  # NOQA pylint: disable=unused-import
    from pycdlib import eltorito  # NOQA pylint: disable=unused-import
    from pycdlib import udf  # NOQA pylint: disable=unused-import


class Inode(object):
    '''
    A class that represents an inode, the pointer to a piece of data
    (not metadata) on an ISO.
    '''
    __slots__ = ('_initialized', 'new_extent_loc', 'orig_extent_loc',
                 'linked_records', 'data_length', 'manage_fp', 'data_fp',
                 'original_data_location', 'fp_offset', 'boot_info_table',
                 'num_udf')

    DATA_ON_ORIGINAL_ISO = 1
    DATA_IN_EXTERNAL_FP = 2

    def __init__(self):
        # type: () -> None
        self.linked_records = []  # type: List[Tuple[Union[eltorito.EltoritoEntry, udf.UDFFileEntry, dr.DirectoryRecord], bool]]
        self._initialized = False
        self.data_length = 0
        self.num_udf = 0
        self.boot_info_table = None  # type: Optional[eltorito.EltoritoBootInfoTable]
        self.new_extent_loc = -1

    def new(self, length, fp, manage_fp, offset):
        # type: (int, Union[BinaryIO, str], bool, int) -> None
        '''
        Initialize a new Inode.

        Parameters:
         None.
        Returns:
         Nothing.
        '''
        if self._initialized:
            raise pycdlibexception.PyCdlibInternalError('Inode is already initialized')

        self.data_length = length

        self.data_fp = fp
        self.manage_fp = manage_fp
        self.fp_offset = offset
        self.original_data_location = self.DATA_IN_EXTERNAL_FP

        self._initialized = True

    def parse(self, extent, length, fp, log_block_size):
        # type: (int, int, BinaryIO, int) -> None
        '''
        Parse an existing Inode.  This just saves off the extent for later use.

        Parameters:
         extent - The original extent that the data lives at.
        Returns:
         Nothing.
        '''
        if self._initialized:
            raise pycdlibexception.PyCdlibInternalError('Inode is already initialized')

        self.orig_extent_loc = extent

        self.data_length = length

        self.data_fp = fp
        self.manage_fp = False
        self.fp_offset = extent * log_block_size
        self.original_data_location = self.DATA_ON_ORIGINAL_ISO

        self._initialized = True

    def extent_location(self):
        # type: () -> int
        '''
        Get the current location of this Inode on the ISO.

        Parameters:
         None.
        Returns:
         The extent location of this Inode on the ISO.
        '''
        if not self._initialized:
            raise pycdlibexception.PyCdlibInternalError('Inode is not initialized')

        if self.new_extent_loc < 0:
            return self.orig_extent_loc
        return self.new_extent_loc

    def set_extent_location(self, extent):
        # type: (int) -> None
        '''
        Set the current location of this Inode on the ISO.

        Parameters:
         extent - The new extent location for this Inode.
        Returns:
         Nothing.
        '''
        if not self._initialized:
            raise pycdlibexception.PyCdlibInternalError('Inode is not initialized')

        self.new_extent_loc = extent

    def get_data_length(self):
        # type: () -> int
        '''
        Get the length of the data pointed to by this Inode.

        Parameters:
         None.
        Returns:
         The length of the data pointed to by this Inode.
        '''
        if not self._initialized:
            raise pycdlibexception.PyCdlibInternalError('Inode is not initialized')

        return self.data_length

    def add_boot_info_table(self, boot_info_table):
        # type: (eltorito.EltoritoBootInfoTable) -> None
        '''
        Add a boot info table to this Inode.

        Parameters:
         boot_info_table - The Boot Info Table object to add to this Inode.
        Returns:
         Nothing.
        '''
        if not self._initialized:
            raise pycdlibexception.PyCdlibInternalError('Inode is not initialized')

        self.boot_info_table = boot_info_table

    def update_fp(self, fp, length):
        # type: (BinaryIO, int) -> None
        '''
        Update the Inode to use a different file object and length.

        Parameters:
         fp - A file object that contains the data for this Inode.
         length - The length of the data.
        Returns:
         Nothing.
        '''
        if not self._initialized:
            raise pycdlibexception.PyCdlibInternalError('Inode is not initialized')

        self.original_data_location = self.DATA_IN_EXTERNAL_FP
        self.data_fp = fp
        self.data_length = length
        self.fp_offset = 0


class InodeOpenData(object):
    '''
    A class to be a contextmanager for opening data on a DirectoryRecord object.
    '''
    __slots__ = ('ino', 'logical_block_size', 'data_fp')

    def __init__(self, ino, logical_block_size):
        # type: (Inode, int) -> None
        self.ino = ino
        self.logical_block_size = logical_block_size

    def __enter__(self):
        if self.ino.manage_fp:
            # In the case that we are managing the FP, the data_fp member
            # actually contains the filename, not the fp.  Use that to
            # our advantage here.
            self.data_fp = open(self.ino.data_fp, 'rb')
        else:
            self.data_fp = self.ino.data_fp

        if self.ino.original_data_location == self.ino.DATA_ON_ORIGINAL_ISO:
            self.data_fp.seek(self.ino.orig_extent_loc * self.logical_block_size)
        else:
            self.data_fp.seek(self.ino.fp_offset)

        return self.data_fp, self.ino.data_length

    def __exit__(self, *args):
        if self.ino.manage_fp:
            self.data_fp.close()



from __future__ import absolute_import

import pytest
import os
import sys
try:
    from cStringIO import StringIO as BytesIO
except ImportError:
    from io import BytesIO
import struct

prefix = '.'
for i in range(0, 3):
    if os.path.isdir(os.path.join(prefix, 'pycdlib')):
        sys.path.insert(0, prefix)
        break
    else:
        prefix = '../' + prefix

import pycdlib.path_table_record

def test_path_table_record_parse_initialized_twice():
    ptr = pycdlib.path_table_record.PathTableRecord()
    ptr.parse(b'\x00\x00\x00\x00\x00\x00\x00\x00')

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        ptr.parse(b'\x00\x00\x00\x00\x00\x00\x00\x00')
    assert(str(excinfo.value) == 'Path Table Record already initialized')

def test_path_table_record_record_little_endian_not_initialized():
    ptr = pycdlib.path_table_record.PathTableRecord()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        ptr.record_little_endian()
    assert(str(excinfo.value) == 'Path Table Record not initialized')

def test_path_table_record_record_big_endian_not_initialized():
    ptr = pycdlib.path_table_record.PathTableRecord()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        ptr.record_big_endian()
    assert(str(excinfo.value) == 'Path Table Record not initialized')

def test_path_table_record_new_root_initialized_twice():
    ptr = pycdlib.path_table_record.PathTableRecord()
    ptr.new_root()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        ptr.new_root()
    assert(str(excinfo.value) == 'Path Table Record already initialized')

def test_path_table_record_new_dir_initialized_twice():
    ptr = pycdlib.path_table_record.PathTableRecord()
    ptr.new_dir(b'foo')

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        ptr.new_dir(b'foo')
    assert(str(excinfo.value) == 'Path Table Record already initialized')

def test_path_table_record_update_extent_location_not_initialized():
    ptr = pycdlib.path_table_record.PathTableRecord()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        ptr.update_extent_location(0)
    assert(str(excinfo.value) == 'Path Table Record not initialized')

def test_path_table_record_update_parent_directory_number_not_initialized():
    ptr = pycdlib.path_table_record.PathTableRecord()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        ptr.update_parent_directory_number(0)
    assert(str(excinfo.value) == 'Path Table Record not initialized')

def test_path_table_record_equal_to_be_not_initialized():
    ptr = pycdlib.path_table_record.PathTableRecord()

    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        ptr.equal_to_be(None)
    assert(str(excinfo.value) == 'Path Table Record not initialized')



from __future__ import absolute_import

import pytest
import os
import sys
try:
    from cStringIO import StringIO as BytesIO
except ImportError:
    from io import BytesIO
import struct
import time

prefix = '.'
for i in range(0, 3):
    if os.path.isdir(os.path.join(prefix, 'pycdlib')):
        sys.path.insert(0, prefix)
        break
    else:
        prefix = '../' + prefix

import pycdlib.utils
import pycdlib.pycdlibexception

def test_swab_32bit():
    assert(pycdlib.utils.swab_32bit(0x89) == 0x89000000)

def test_swab_32bit_bad_input():
    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        pycdlib.utils.swab_32bit(-1)
    assert(str(excinfo.value) == 'Invalid integer passed to swab; must be unsigned 32-bits!')

def test_swab_16bit():
    assert(pycdlib.utils.swab_16bit(0x55aa) == 0xaa55)

def test_swab_16bit_bad_input():
    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInternalError) as excinfo:
        pycdlib.utils.swab_16bit(-1)
    assert(str(excinfo.value) == 'Invalid integer passed to swab; must be unsigned 16-bits!')

def test_ceiling_div():
    assert(pycdlib.utils.ceiling_div(0, 2048) == 0)

def test_ceiling_div2():
    assert(pycdlib.utils.ceiling_div(2048, 2048) == 1)

def test_ceiling_div3():
    assert(pycdlib.utils.ceiling_div(2049, 2048) == 2)

def test_ceiling_div_nan():
    with pytest.raises(ZeroDivisionError) as exc_info:
        pycdlib.utils.ceiling_div(2048, 0)

def test_copy_data():
    infp = BytesIO()
    outfp = BytesIO()

    infp.write(b'\x00'*1)
    infp.seek(0)

    pycdlib.utils.copy_data(1, 8192, infp, outfp)

    assert(outfp.getvalue() == b'\x00')

def test_copy_data_short():
    infp = BytesIO()
    outfp = BytesIO()

    infp.write(b'\x00'*10)
    infp.seek(0)

    pycdlib.utils.copy_data(100, 8192, infp, outfp)

    assert(outfp.getvalue() == b'\x00'*10)

def test_encode_space_pad_too_short():
    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInvalidInput) as exc_info:
        pycdlib.utils.encode_space_pad(b'hello', 4, 'ascii')
    assert(str(exc_info.value) == 'Input string too long!')

def test_encode_space_pad_no_pad():
    assert(pycdlib.utils.encode_space_pad(b'hello', 5, 'ascii') == b'hello')

def test_encode_space_pad_one():
    assert(pycdlib.utils.encode_space_pad(b'hello', 6, 'ascii') == b'hello ')

def test_encode_space_pad_extra():
    assert(pycdlib.utils.encode_space_pad(b'hello', 11, 'utf-16_be') == b'\x00h\x00e\x00l\x00l\x00o\x00')

def test_normpath_double_slashes_beginning():
    assert(pycdlib.utils.normpath('//') == b'/')

def test_normpath_double_slashes_middle():
    assert(pycdlib.utils.normpath('/foo//bar') == b'/foo/bar')

def test_normpath_with_dotdot():
    assert(pycdlib.utils.normpath('/foo/bar/../baz') == b'/foo/baz')

def test_normpath_with_dotdot_after_slash():
    assert(pycdlib.utils.normpath('/../foo') == b'/foo')

def test_normpath_empty():
    assert(pycdlib.utils.normpath('') == b'.')

def save_and_set_tz(newtz):
    if 'TZ' in os.environ:
        oldtz = os.environ['TZ']
    else:
        oldtz = None

    os.environ['TZ'] = newtz
    time.tzset()

    return oldtz

def restore_tz(oldtz):
    if oldtz is not None:
        os.environ['TZ'] = oldtz
    else:
        del os.environ['TZ']
    time.tzset()

def test_gmtoffset_from_tm():
    oldtz = save_and_set_tz('US/Eastern')
    now = 1546914300.0
    assert(pycdlib.utils.gmtoffset_from_tm(now, time.localtime(now)) == -20)
    restore_tz(oldtz)

def test_gmtoffset_from_tm_day_rollover():
    # Setup the timezone to Tokyo
    oldtz = save_and_set_tz('Asia/Tokyo')

    # This tm is carefully chosen so that the day of the week is the next day
    # in the Tokyo region.
    now = 1550417871
    local = time.localtime(now)
    assert(pycdlib.utils.gmtoffset_from_tm(now, local) == 36)

    restore_tz(oldtz)

def test_zero_pad():
    fp = BytesIO()
    pycdlib.utils.zero_pad(fp, 5, 10)
    assert(fp.getvalue() == b'\x00'*5)

def test_zero_pad_no_pad():
    fp = BytesIO()
    pycdlib.utils.zero_pad(fp, 5, 5)
    assert(fp.getvalue() == b'')

def test_zero_pad_negative_pad():
    fp = BytesIO()
    pycdlib.utils.zero_pad(fp, 5, 4)
    assert(fp.getvalue() == b'\x00'*3)

def test_starts_with_slash():
    assert(pycdlib.utils.starts_with_slash(b'/'))

def test_starts_with_slash_no_slash():
    assert(not pycdlib.utils.starts_with_slash(b'foo/bar'))

def test_starts_with_slash_double_slash():
    assert(pycdlib.utils.starts_with_slash(b'//foo/bar'))

def test_split_path():
    assert(pycdlib.utils.split_path(b'/foo/bar') == [b'foo', b'bar'])

def test_split_path_no_leading_slash():
    with pytest.raises(pycdlib.pycdlibexception.PyCdlibInvalidInput) as exc_info:
        pycdlib.utils.split_path(b'foo/bar')
    assert(str(exc_info.value) == 'Must be a path starting with /')

def test_split_path_single():
    assert(pycdlib.utils.split_path(b'/foo') == [b'foo'])

def test_split_path_slash_only():
    assert(pycdlib.utils.split_path(b'/') == [b''])

def test_split_path_trailing_slash():
    assert(pycdlib.utils.split_path(b'/foo/') == [b'foo', b''])

def test_file_object_supports_binary_bytesio():
    fp = BytesIO()
    assert(pycdlib.utils.file_object_supports_binary(fp))

def test_truncate_basename_isolevel4():
    assert(pycdlib.utils.truncate_basename('foo', 4, False) == 'foo')

def test_truncate_basename_isolevel3():
    assert(pycdlib.utils.truncate_basename('foo', 3, False) == 'FOO')

def test_mangle_file_for_iso9660_isolevel4_no_ext():
    assert(pycdlib.utils.mangle_file_for_iso9660('foo', 4) == ('foo', ''))

def test_mangle_file_for_iso9660_isolevel4_with_ext():
    assert(pycdlib.utils.mangle_file_for_iso9660('foo.txt', 4) == ('foo', 'txt'))

def test_mangle_file_for_iso9660_isolevel3_with_empty_ext():
    assert(pycdlib.utils.mangle_file_for_iso9660('foo.', 3) == ('FOO_', ';1'))

def test_file_object_supports_binary_real_file(tmpdir):
    testout = tmpdir.join('foo')
    with open(str(testout), 'wb') as outfp:
        assert(pycdlib.utils.file_object_supports_binary(outfp))

def test_file_object_does_not_support_binary_real_file(tmpdir):
    testout = tmpdir.join('foo')
    with open(str(testout), 'w') as outfp:
        assert(not pycdlib.utils.file_object_supports_binary(outfp))



# Copyright (C) 2015-2019  Chris Lalancette <clalancette@gmail.com>

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

'''
Classes and utilities for ISO date support.
'''

from __future__ import absolute_import

import struct
import time
try:
    from functools import lru_cache
except ImportError:
    from pycdlib.backport_functools import lru_cache  # type: ignore

from pycdlib import pycdlibexception
from pycdlib import utils


@lru_cache(maxsize=256)
def string_to_timestruct(input_string):
    # type: (bytes) -> time.struct_time
    '''
    A cacheable function to take an input string and decode it into a
    time.struct_time from the time module.  If the string cannot be decoded
    because of an illegal value, then the all-zero time.struct_time will be
    returned instead.

    Parameters:
     input_string - The string to attempt to parse.
    Returns:
     A time.struct_time object representing the time.
    '''
    try:
        timestruct = time.strptime(input_string.decode('utf-8'), VolumeDescriptorDate.TIME_FMT)
    except ValueError:
        # Ecma-119, 8.4.26.1 specifies that if the string was all the digit
        # zero, with the last byte 0, the time wasn't specified.  In that
        # case, time.strptime() with our format will raise a ValueError.
        # In practice we have found that some ISOs specify various wacky
        # things in this field, so if we see *any* ValueError, we just
        # assume the date is unspecified and go with that.
        timestruct = time.struct_time((0, 0, 0, 0, 0, 0, 0, 0, 0))

    return timestruct


class DirectoryRecordDate(object):
    '''
    A class to represent a Directory Record date as described in Ecma-119
    section 9.1.5.  The Directory Record date consists of the number of years
    since 1900, the month, the day of the month, the hour, the minute, the
    second, and the offset from GMT in 15 minute intervals.  There are two main
    ways to use this class: either to instantiate and then parse a string to
    fill in the fields (the parse() method), or to create a new entry with a
    tm structure (the new() method).
    '''
    FMT = '=BBBBBBb'

    __slots__ = ('_initialized', 'years_since_1900', 'month', 'day_of_month',
                 'hour', 'minute', 'second', 'gmtoffset')

    def __init__(self):
        # type: () -> None
        self._initialized = False

    def parse(self, datestr):
        # type: (bytes) -> None
        '''
        Parse a Directory Record date out of a string.

        Parameters:
         datestr - The string to parse the date out of.
        Returns:
         Nothing.
        '''
        if self._initialized:
            raise pycdlibexception.PyCdlibInternalError('Directory Record Date already initialized')

        (self.years_since_1900, self.month, self.day_of_month, self.hour,
         self.minute, self.second,
         self.gmtoffset) = struct.unpack_from(self.FMT, datestr, 0)

        self._initialized = True

    def new(self):
        # type: () -> None
        '''
        Create a new Directory Record date based on the current time.

        Parameters:
         tm - An optional argument that must be None
        Returns:
         Nothing.
        '''
        if self._initialized:
            raise pycdlibexception.PyCdlibInternalError('Directory Record Date already initialized')

        # This algorithm was ported from cdrkit, genisoimage.c:iso9660_date()
        tm = time.time()
        local = time.localtime(tm)
        self.years_since_1900 = local.tm_year - 1900
        self.month = local.tm_mon
        self.day_of_month = local.tm_mday
        self.hour = local.tm_hour
        self.minute = local.tm_min
        self.second = local.tm_sec
        self.gmtoffset = utils.gmtoffset_from_tm(tm, local)
        self._initialized = True

    def record(self):
        # type: () -> bytes
        '''
        Return a string representation of the Directory Record date.

        Parameters:
         None.
        Returns:
         A string representing this Directory Record Date.
        '''
        if not self._initialized:
            raise pycdlibexception.PyCdlibInternalError('Directory Record Date not initialized')

        return struct.pack(self.FMT, self.years_since_1900, self.month,
                           self.day_of_month, self.hour, self.minute,
                           self.second, self.gmtoffset)

    def __ne__(self, other):
        return self.years_since_1900 != other.years_since_1900 or self.month != other.month or self.day_of_month != other.day_of_month or self.hour != other.hour or self.minute != other.minute or self.second != other.second or self.gmtoffset != other.gmtoffset


class VolumeDescriptorDate(object):
    '''
    A class to represent a Volume Descriptor Date as described in Ecma-119
    section 8.4.26.1.  The Volume Descriptor Date consists of a year (from 1 to
    9999), month (from 1 to 12), day of month (from 1 to 31), hour (from 0
    to 23), minute (from 0 to 59), second (from 0 to 59), hundredths of second,
    and offset from GMT in 15-minute intervals (from -48 to +52) fields.  There
    are two main ways to use this class: either to instantiate and then parse a
    string to fill in the fields (the parse() method), or to create a new entry
    with a tm structure (the new() method).
    '''

    TIME_FMT = '%Y%m%d%H%M%S'

    EMPTY_STRING = b'0' * 16 + b'\x00'

    __slots__ = ('_initialized', 'year', 'month', 'dayofmonth', 'hour',
                 'minute', 'second', 'hundredthsofsecond', 'gmtoffset',
                 'date_str')

    def __init__(self):
        # type: () -> None
        self._initialized = False

    def parse(self, datestr):
        # type: (bytes) -> None
        '''
        Parse a Volume Descriptor Date out of a string.  A string of all zeros
        is valid, which means that the date in this field was not specified.

        Parameters:
          datestr - string to be parsed
        Returns:
          Nothing.
        '''
        if self._initialized:
            raise pycdlibexception.PyCdlibInternalError('This Volume Descriptor Date object is already initialized')

        if len(datestr) != 17:
            raise pycdlibexception.PyCdlibInvalidISO('Invalid ISO9660 date string')

        timestruct = string_to_timestruct(datestr[:-3])
        self.year = timestruct.tm_year
        self.month = timestruct.tm_mon
        self.dayofmonth = timestruct.tm_mday
        self.hour = timestruct.tm_hour
        self.minute = timestruct.tm_min
        self.second = timestruct.tm_sec
        if timestruct.tm_year == 0 and timestruct.tm_mon == 0 and timestruct.tm_mday == 0 and timestruct.tm_hour == 0 and timestruct.tm_min == 0 and timestruct.tm_sec == 0:
            self.hundredthsofsecond = 0
            self.gmtoffset = 0
            self.date_str = self.EMPTY_STRING
        else:
            try:
                self.hundredthsofsecond = int(datestr[14:16])
            except ValueError:
                # We've seen ISOs in the wild (made by MagicISO) that fill
                # hundredthsofseconds with b'\x00\x00'.  Handle that here.
                self.hundredthsofsecond, = struct.unpack('>H', datestr[14:16])
            self.gmtoffset, = struct.unpack_from('=b', datestr, 16)
            self.date_str = datestr

        self._initialized = True

    def record(self):
        # type: () -> bytes
        '''
        Return the date string for this object.

        Parameters:
          None.
        Returns:
          Date as a string.
        '''
        if not self._initialized:
            raise pycdlibexception.PyCdlibInternalError('This Volume Descriptor Date is not initialized')

        return self.date_str

    def new(self, tm=0.0):
        # type: (float) -> None
        '''
        Create a new Volume Descriptor Date.  If tm is None, then this Volume
        Descriptor Date will be full of zeros (meaning not specified).  If tm
        is not None, it is expected to be a struct_time object, at which point
        this Volume Descriptor Date object will be filled in with data from that
        struct_time.

        Parameters:
          tm - struct_time object to base new VolumeDescriptorDate off of,
               or 0.0 for an empty VolumeDescriptorDate.
        Returns:
          Nothing.
        '''
        if self._initialized:
            raise pycdlibexception.PyCdlibInternalError('This Volume Descriptor Date object is already initialized')

        if tm != 0.0:
            local = time.localtime(tm)
            self.year = local.tm_year
            self.month = local.tm_mon
            self.dayofmonth = local.tm_mday
            self.hour = local.tm_hour
            self.minute = local.tm_min
            self.second = local.tm_sec
            self.hundredthsofsecond = 0
            self.gmtoffset = utils.gmtoffset_from_tm(tm, local)
            self.date_str = time.strftime(self.TIME_FMT, local).encode('utf-8') + '{:0<2}'.format(self.hundredthsofsecond).encode('utf-8') + struct.pack('=b', self.gmtoffset)
        else:
            self.year = 0
            self.month = 0
            self.dayofmonth = 0
            self.hour = 0
            self.minute = 0
            self.second = 0
            self.hundredthsofsecond = 0
            self.gmtoffset = 0
            self.date_str = self.EMPTY_STRING

        self._initialized = True

    def __ne__(self, other):
        return self.year != other.year or self.month != other.month or self.dayofmonth != other.dayofmonth or self.hour != other.hour or self.minute != other.minute or self.second != other.second or self.hundredthsofsecond != other.hundredthsofsecond or self.gmtoffset != other.gmtoffset or self.date_str != other.date_str
